/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.composite.c.controllers;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.ComponentVisitor;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.cecilia.adl.AbstractInstantiationTask;
import org.objectweb.fractal.cecilia.adl.InstanceNameProvider;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.TypeNameProvider;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.task.core.AbstractTaskFactoryUser;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ClientInterfaceForEach;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Visitor component that builds the instantiation of the content controller
 * data structure of a default composite component.
 */
public class CCDynamicDescriptorVisitor extends AbstractTaskFactoryUser
    implements
      ComponentVisitor {

  // ---------------------------------------------------------------------------
  // Implementation of abstract methods of AbstractClientInterfaceVisitor
  // ---------------------------------------------------------------------------

  /**
   * Visits {@link ComponentContainer} nodes and creates a task that writes the
   * source code containing the initialization of the content controller data
   * structure.
   */
  public org.objectweb.fractal.api.Component visit(final List<Node> path,
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException, TaskException {
    final Component[] subComponents = container.getComponents();
    return taskFactoryItf.newPrimitiveTask(new CCInstantiationTask(
        subComponents), container, Arrays.asList(subComponents));
  }

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  /**
   * Builds source code for the instantiation of the content controller data
   * structure for the given component node. This task provides the produced
   * source code.
   */
  @TaskParameters({"componentNode", "subComponents"})
  @ServerInterfaces(@ServerInterface(name = "cc-instantiation-provider", signature = SourceCodeProvider.class, record = "role:controllerInstantiation, id:%, controller:CC", parameters = "componentNode"))
  public static class CCInstantiationTask extends AbstractInstantiationTask {

    protected final Component[]                       subComponents;

    // -------------------------------------------------------------------------
    // Task client interfaces
    // -------------------------------------------------------------------------

    /** The {@link TypeNameProvider} task for each sub component. */
    @ClientInterfaceForEach(prefix = "subComp-type-names", signature = TypeNameProvider.class, iterable = "subComponents", record = "role:subComponentTypeNameProvider, id:%", parameters = "subComponents.element")
    public final Map<Component, TypeNameProvider>     subComponentTypeNameProviderItfs       = new HashMap<Component, TypeNameProvider>();

    /**
     * The {@link SourceCodeProvider} task that provides definition of component
     * type for each sub component.
     */
    @ClientInterfaceForEach(prefix = "subComp-type-definitions", signature = SourceCodeProvider.class, iterable = "subComponents", record = "role:subComponentTypeDefinitionProvider, id:%", parameters = "subComponents.element")
    public final Map<Component, SourceCodeProvider>   subComponentTypeDefinitionProviderItfs = new HashMap<Component, SourceCodeProvider>();

    /** The {@link InstanceNameProvider} task for each sub component. */
    @ClientInterfaceForEach(prefix = "subComp-instance-names", signature = InstanceNameProvider.class, iterable = "subComponents", record = "role:subComponentInstanceNameProvider, id:%", parameters = "subComponents.element")
    public final Map<Component, InstanceNameProvider> subComponentInstanceNameProviderItfs   = new HashMap<Component, InstanceNameProvider>();

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param subComponents the sub component nodes.
     */
    public CCInstantiationTask(final Component[] subComponents) {
      this.subComponents = subComponents;
    }

    // -------------------------------------------------------------------------
    // Implementation of abstract methods of AbstractInstantiationTask
    // -------------------------------------------------------------------------

    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter("Content Controller Data Descriptor");

      final String instanceName = instanceNameProviderItf.getCInstanceName();

      final String[] subCompInstanceNames = new String[subComponents.length];
      final String[] subCompNames = new String[subComponents.length];
      for (int i = 0; i < subComponents.length; i++) {
        subCompInstanceNames[i] = subComponentInstanceNameProviderItfs.get(
            subComponents[i]).getCInstanceName();
        final int idx = subComponents[i].toString().lastIndexOf('.');

        subCompNames[i] = (idx < 0) ? subComponents[i].getName().replace('-',
            '_') : subComponents[i].getName().replace('-', '_').substring(idx);

      }

      cw.append("/* extern reference for the description of ").append(
          instanceName).append("'s sub component*/").endl();
      for (final String subCompInstanceName : subCompInstanceNames) {
        cw.append("extern component_desc_t ").append(subCompInstanceName)
            .append("_desc;").endl();
      }

      cw.append("/* Defining the ").append(instanceName).append(
          "'s sub component index numbers*/").endl();
      cw.append("#define thisComposite_exp 0xff").endl();
      for (int i = 0; i < subCompInstanceNames.length; i++) {
        cw.append("#define ").append(
            subComponents[i].getName().replace('-', '_')).append("_exp ")
            .append(i).endl();
      }

      cw.append("/* sub component table for the composite ").append(
          instanceName).append(" */").endl();
      cw.append("static sub_component_desc_t ").append(instanceName).append(
          "_sub_components_desc[] = {").endl();
      for (int i = 0; i < subComponents.length; i++) {
        // This part is probably not correct because it does not handle
        // composite sub components
        cw.append("{\"").append(subCompNames[i]).append("\", &").append(
            subCompInstanceNames[i]).append("_desc },").endl();
      }
      cw.appendln("};");

      return cw.toString();
    }
  }
}
