/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.thinkMC.components;

import static org.objectweb.fractal.api.type.TypeFactory.OPTIONAL;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.ComponentVisitor;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.cecilia.adl.AbstractInstantiationTask;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.task.core.AbstractTaskFactoryUser;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ClientInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Builds the description structure of a component written in
 * <code>thinkMC</code> language.
 */
public class ComponentDescriptorVisitor extends AbstractTaskFactoryUser
    implements
      ComponentVisitor {

// ---------------------------------------------------------------------------
// Implementation of the ComponentVisitor interface
// ---------------------------------------------------------------------------

  /**
   * Visits {@link ComponentContainer} nodes and creates a task that writes the
   * source code containing the component data structure definitions.
   */
  public Component visit(final List<Node> path,
      final ComponentContainer container, final Map<Object, Object> context)
      throws ADLException, TaskException {
    return taskFactoryItf.newPrimitiveTask(new ComponentDefinitionTask(),
        container);
  }

// ---------------------------------------------------------------------------
// Task classes
// ---------------------------------------------------------------------------

  /**
   * Builds source code for the instantiation/initialization of the component
   * data structure for the given component node. This task provides the
   * produced source code. It uses the source code provided for
   * imported/exported interfaces, controllers, and the attributes of the
   * component.
   */
  @TaskParameters("componentNode")
  @ServerInterfaces(@ServerInterface(name = "component-instantiation-provider", signature = SourceCodeProvider.class, record = "role:componentInstantiation, id:%", parameters = "componentNode"))
  public static class ComponentDefinitionTask extends AbstractInstantiationTask {

// -------------------------------------------------------------------------
// Task client interfaces
// -------------------------------------------------------------------------

    /** Source code for the initialization of interface instances. */
    @ClientInterface(name = "client-interface-instance-initialization-provider", contingency = OPTIONAL, record = "role:clientInterfaceInstantiation, id:%, codePiece:initialization", parameters = "componentNode")
    public SourceCodeProvider                    clientInterfaceInstancesProviderInitializationItf;

    /** Source code for the initialization of interface instances. */
    @ClientInterface(name = "server-interface-instance-initialization-provider", contingency = OPTIONAL, record = "role:serverInterfaceInstantiation, id:%, codePiece:initialization", parameters = "componentNode")
    public SourceCodeProvider                    serverInterfaceInstancesProviderInitializationItf;

    /** Source code containing includes necessary for the interface instances. */
    @ClientInterface(name = "interface-instance-include-provider", contingency = OPTIONAL, record = "role:interfaceInstantiation, id:%", parameters = "componentNode")
    public SourceCodeProvider                    interfaceInstancesProviderIncludeItf;

    /** Source code for attribute definitions. */
    @ClientInterface(name = "attribute-instantiation-provider", contingency = OPTIONAL, record = "role:attributeInstantiation, id:%", parameters = "componentNode")
    public SourceCodeProvider                    attributeInstantiationProviderItf;

    /** Source codes for each controller definition. */
    @ClientInterface(name = "controller-instantiation-provider", signature = SourceCodeProvider.class, record = "role:controllerInstantiation, id:%", parameters = "componentNode")
    public final Map<String, SourceCodeProvider> controllerInstantiationsProviderItf = new HashMap<String, SourceCodeProvider>();

// -------------------------------------------------------------------------
// Implementation of abstract methods of AbstractInstantiationTask
// -------------------------------------------------------------------------
    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter("Primitive Description Builder");
      final String definitionName = typeNameProviderItf.getCTypeName();
      final String instanceName = instanceNameProviderItf.getCInstanceName();

      /* Include descriptor structure declaration */
      cw
          .appendln("#include \"minus/common/generic/fractal/api/componentDesc.h\"");

      /* Add comments containing instance name and type */
      cw.appendln("// --------------------------------------------------------"
          + "---------------------");
      cw.append("// Declare instance ").append(
          instanceNameProviderItf.getInstanceName()).endl();
      cw.append("//        from type ").append(
          typeNameProviderItf.getTypeName()).endl();
      cw.appendln("// --------------------------------------------------------"
          + "---------------------");

      if (interfaceInstancesProviderIncludeItf != null)
        cw.appendln(interfaceInstancesProviderIncludeItf.getSourceCode());

      if (attributeInstantiationProviderItf != null) {
        cw.appendln(attributeInstantiationProviderItf.getSourceCode());
      }

      cw.append("primitive_desc_t ").append(instanceName).append("_desc = {")
          .endl();
      cw.append("PRIMITIVE_TYPE, // type").endl();
      cw.append("0, // properties").endl();
      cw.append("sizeof(struct ").append(definitionName).append(
          "_t), // dataSize").endl();

      if (serverInterfaceInstancesProviderInitializationItf != null) {
        final String exporteds = serverInterfaceInstancesProviderInitializationItf
            .getSourceCode();
        cw.append("sizeof(").append(exporteds).append(") / ").append(
            "sizeof(provided_interface_desc_t), //nbProvidedItfs").endl();
        cw.append(exporteds).append(", //providedItfs").endl();
      } else {
        cw.append("0, //nbProvidedItfs").endl();
        cw.append("0, //providedItfs").endl();
      }

      if (clientInterfaceInstancesProviderInitializationItf != null) {
        final String importeds = clientInterfaceInstancesProviderInitializationItf
            .getSourceCode();
        cw.append("sizeof(").append(importeds).append(") / ").append(
            "sizeof(required_interface_desc_t), //nbRequiredItfs").endl();
        cw.append(importeds).append(", // requiredItfs").endl();
      } else {
        cw.append("0, //nbRequiredItfs").endl();
        cw.append("0, //requiredItfs").endl();
      }

      if (attributeInstantiationProviderItf != null) {
        cw.append("sizeof(").append(instanceName)
            .append("_attributes_desc) / ").append(
                "sizeof(attribute_desc_t), //nbAttribute").endl();
        cw.append(instanceName).append("_attributes_desc // attributes").endl();
      } else {
        cw.appendln("0, //nbAttribute");
        cw.appendln("0 // attributes");
      }
      cw.appendln("};");

      return cw.toString();
    }
  }

}
