/***
 * Think ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: think@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.thinkMC.types;

import static java.util.Arrays.asList;
import static org.objectweb.fractal.task.core.TaskFactory.EXPORT_ALL;

import java.util.LinkedHashMap;
import java.util.List;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.cecilia.adl.AbstractInstantiationTask;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;
import org.objectweb.fractal.cecilia.adl.interfaces.AbstractServerInterfaceVisitor;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Visitor component that builds the descriptor of the server interfaces of the
 * component for <code>ThinkMC</code> dialect.
 */
public class ServerInterfaceDescriptorVisitor
    extends
      AbstractServerInterfaceVisitor {

  // ---------------------------------------------------------------------------
  // Implementation of abstract methods of AbstractServerInterfaceVisitor
  // ---------------------------------------------------------------------------

  /**
   * Creates a task that writes the source code containing the instantiation of
   * the client interfaces of the component.
   */
  @Override
  protected Component createTask(final ComponentContainer container,
      final List<TypeInterface> serverInterfaces) throws ADLException,
      TaskException {

    final Component includeTask = createIncludeServerInstanceTask(container,
        serverInterfaces);
    final Component initializationTask = createInitializationServerInstanceTask(
        container, serverInterfaces);

    return taskFactoryItf.newCompositeTask(asList(includeTask,
        initializationTask), EXPORT_ALL, null);
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  protected Component createIncludeServerInstanceTask(
      final ComponentContainer container,
      final List<TypeInterface> serverInterfaces) throws TaskException {
    return taskFactoryItf.newPrimitiveTask(new IncludeServerInstanceTask(
        serverInterfaces), container);
  }

  protected Component createInitializationServerInstanceTask(
      final ComponentContainer container,
      final List<TypeInterface> serverInterfaces) throws TaskException {
    return taskFactoryItf.newPrimitiveTask(
        new InitializationServerInstanceTask(serverInterfaces), container);
  }

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  /**
   * Builds instantiation source code for server interfaces for the given
   * component node. This task provides the produced source code.
   */
  @TaskParameters({"componentNode"})
  @ServerInterfaces(@ServerInterface(name = "source-code-provider", signature = SourceCodeProvider.class, record = "role:serverInterfaceInstantiation, id:%, codePiece:include", parameters = "componentNode"))
  public static class IncludeServerInstanceTask
      extends
        AbstractInstantiationTask {

    protected final List<TypeInterface> exportedItfNodes;

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param exportedItfNodes the list of exported interface nodes belonging to
     *            the component.
     */
    public IncludeServerInstanceTask(final List<TypeInterface> exportedItfNodes) {
      // The order of the interfaces is important, this is why an array of
      // exportedIrfNodes is used for iterating of the exported interfaces
      // rather than iterating directly on the set of idlHeaderItfs client
      // interfaces.
      this.exportedItfNodes = exportedItfNodes;
    }

    // -------------------------------------------------------------------------
    // Implementation of abstract methods of AbstractInstantiationTask
    // -------------------------------------------------------------------------

    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter("Server Interfaces Descriptor");

      final String name = typeNameProviderItf.getCTypeName();
      final String structName = name + "_exporteds_desc";

      // Build provided_interface_desc_t structures
      final StringBuilder offset = new StringBuilder();
      offset.append(name.toUpperCase()).append("_EXPORTEDS_BASE_OFFSET");

      cw.append("static provided_interface_desc_t ").append(structName).append(
          "[] = {").endl();
      for (final TypeInterface itf : exportedItfNodes) {
        cw.append("{").endl();
        cw.append("\"").append(itf.getName()).append("\",  // name").endl();
        cw.append("&(").append(name).append("_").append(
            itf.getName().replace('-', '_')).append("meth),  // vtable").endl();
        cw.append("(").append(offset.toString()).append(")").endl();

        /* Add the signature */
        cw.append("#if defined(TYPEDINTERFACE)").endl();
        cw.append(", \"").append(itf.getSignature()).append("\"").endl();
        cw.append("#endif").endl();

        cw.append("},").endl();

        offset.append("\n + sizeof(R").append(
            itf.getSignature().replace('.', '_')).append(")");
      }
      cw.append("};").endl().endl();

      return cw.toString();
    }
  }

  /**
   * Builds part of the source code for imported interface instantiation for the
   * given component node. More precisely, this task builds the source codes
   * that initializes the client interfaces to point to the server interfaces
   * they are statically bound to. This task provides the produced source code.
   * <br>
   * The <code>serverComponents</code> arguments must refer to the collection
   * of the server components to which the given component is statically bound
   * to.
   */
  @ServerInterfaces(@ServerInterface(name = "source-code-provider", signature = SourceCodeProvider.class, record = "role:serverInterfaceInstantiation, id:%, codePiece:initialization", parameters = "componentNode"))
  public class InitializationServerInstanceTask
      extends
        IncludeServerInstanceTask {

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @see IncludeServerInstanceTask#IncludeServerInstanceTask(LinkedHashMap)
     */
    public InitializationServerInstanceTask(
        final List<TypeInterface> ServerInterfaces) {
      super(ServerInterfaces);
    }

    @Override
    protected String processSourceCode() throws Exception {
      final CodeWriter cw = new CodeWriter(
          "Server Interfaces Descriptor Builder (initialization)");

      final String name = typeNameProviderItf.getCTypeName();
      final String structName = name + "_exporteds_desc";

      cw.append(structName);

      return cw.toString();
    }
  }
}
